from rekall.domain_interval_collection import DomainIntervalCollection
from rekall.interval_set_3d import Interval3D, IntervalSet3D
from rekall.temporal_predicates import overlaps
from rekall.merge_ops import payload_first
import rekall.interval_set_3d_utils as utils
import unittest

from pstats import Stats
import cProfile

from rekall.video_interval_collection import VideoIntervalCollection
from rekall.interval_list import Interval, IntervalList

class TestDomainIntervalCollection(unittest.TestCase):
    @staticmethod
    def get_collection():
        NUM_INTERVALS=150
        NUM_VIDEOS=100
        def get_intervals_for_video(vid):
            intervals = [
                Interval3D((t, t+1),(0,0.5),(0.5,1),vid)
                for t in range(vid, vid+NUM_INTERVALS)]
            return IntervalSet3D(intervals)
        return DomainIntervalCollection(
                {vid: get_intervals_for_video(vid) for vid in range(
                    NUM_VIDEOS)})

    def assertCollectionEq(self, c1,c2):
        self.assertEqual(c1.keys(), c2.keys())
        for key in c1:
            is1 = c1[key]
            is2 = c2[key]
            list1 = [(i.t, i.x, i.y, i.payload) for i in is1.get_intervals()]
            list2 = [(i.t, i.x, i.y, i.payload) for i in is2.get_intervals()]
            self.assertEqual(list1, list2)

    def test_self_merge(self):
        c = TestDomainIntervalCollection.get_collection()
        c2 = c.merge(c, utils.T(overlaps()),
                payload_first, time_window=10)
        self.assertCollectionEq(c2, c)

    def test_filter_size(self):
        c = TestDomainIntervalCollection.get_collection()
        c2 = c.filter_size(min_size=1,max_size=2)
        self.assertCollectionEq(c2, c)

    def test_fold(self):
        c = TestDomainIntervalCollection.get_collection()
        vids = c.keys()
        d = c.fold(lambda m, i: max(m, i.payload), -1)
        self.assertEqual(d, {v: v for v in vids})

    def test_fold_to_set(self):
        c = TestDomainIntervalCollection.get_collection()
        d = c.fold_to_set(lambda acc, i: acc + [i], [])
        self.assertCollectionEq(c, d)

    def test_fold_modify_accumulator_in_place(self):
        def update(acc, i):
            acc.append(i)
            return acc
        c = TestDomainIntervalCollection.get_collection()
        d = c.fold_to_set(update, [])
        self.assertCollectionEq(c, d)

    def test_union(self):
        c= TestDomainIntervalCollection.get_collection()
        c1 = DomainIntervalCollection({v: c[v] for v in c if v % 2 ==0})
        c2 = DomainIntervalCollection({v: c[v] for v in c if v % 2 ==1})
        c3 = c1.union(c2)
        self.assertCollectionEq(c3, c)

    def test_minus(self):
        c= TestDomainIntervalCollection.get_collection()
        c1 = DomainIntervalCollection({v: c[v] for v in c if v % 2 ==0})
        c2 = DomainIntervalCollection({v: c[v] for v in c if v % 2 ==1})
        c3 = c.minus(c1.map(Interval3D.expand_to_frame))
        self.assertCollectionEq(c3,c2)

    def test_collect_by_interval(self):
        c = TestDomainIntervalCollection.get_collection()
        d = DomainIntervalCollection({1: IntervalSet3D([
                Interval3D((t,t)) for t in range(1, 100)])})
        e = c.collect_by_interval(d, utils.T(overlaps()),
                filter_empty=False, time_window=0)
        self.assertEqual(e.keys(), c.keys())



class TestVideoIntervalCollection(unittest.TestCase):
    def setUpProfiler(self):
        self.pr = cProfile.Profile()
        self.pr.enable()

    def tearDownProfiler(self):
        p = Stats(self.pr)
        p.strip_dirs()
        p.sort_stats('cumtime')
        # p.print_stats()

    @staticmethod
    def get_collection():
        NUM_INTERVALS=1500
        NUM_VIDEOS=100
        def get_intervals_for_video(vid):
            intervals = [
                Interval(t, t+1,vid)
                for t in range(vid, vid+NUM_INTERVALS)]
            return intervals
        return VideoIntervalCollection(
                {vid: get_intervals_for_video(vid) for vid in range(
                    NUM_VIDEOS)})

    def assertCollectionEq(self, c1,c2):
        map1 = c1.get_allintervals()
        map2 = c2.get_allintervals()
        self.assertEqual(map1.keys(), map2.keys())
        for key in map1:
            is1 = map1[key]
            is2 = map2[key]
            list1 = [(i.start, i.end, i.payload) for i in is1.intrvls]
            list2 = [(i.start, i.end, i.payload) for i in is2.intrvls]
            self.assertEqual(list1, list2)

    def test_self_merge(self):
        self.setUpProfiler()
        c = TestVideoIntervalCollection.get_collection()
        c2 = c.merge(c, overlaps(), payload_first, working_window=10)
        self.assertCollectionEq(c2, c)
        self.tearDownProfiler()



"""An Extensible Parallel Runtime Library for Rekall.

Runtime: the entrypoint for running large tasks.
    Its construction takes a factory function for a worker pool, which are
    described in the following section.

WorkerPool Factories:
    inline_pool_factory: A factory for a pool that executes tasks in sequence
        in the main process.
    get_spawned_process_pool_factory: Returns a factory function for a pool
        that creates worker processes by spawning new python interpreters.
        The worker processes do not inherit any context from the main process.
    get_forked_process_pool_factory: Returns a factory function for a pool that
        creates worker processes by forking. The worker processes thus inherits
        all the global context from the main process such as global variables.
        However, safely forking a multithreaded program is problematic.

Combiners for Runtime:
    Runtime.run() is a MapReduce routine where a large task is divided into
    smaller chunks and each chunk is fed to a worker (the map step).
    A combiner is responsible for merging the results of each chunk, i.e. the 
    reduction step. The provided combiners are the following.

    union_combiner: The default combiner. It calls `union` method on the
        per-chunk results to merge them, assuming the results have type
        IntervalSet3D or DomainIntervalCollection.
    disjoint_domain_combiner: A faster combiner than union_combiner which
        assumes that the results are of type DomainIntervalCollection and
        every chunk produces its own set of domain keys that are disjoint
        from results of other chunks.

WorkerPool classes:
    Besides using the provided factory functions, one can create their own 
    factory by using the provided WorkerPool classes, or even write their own
    implementation of the WorkerPool interface, which are described below.

    InlineSingleProcessPool: This worker pool uses the main process to
        execute tasks in sequence.
    ForkedProcessPool: This pool uses forking to create worker processes.
    SpawnedProcessPool: This pool spawns fresh python interpreter processes.
        It can run custom initializers when creating the child processes.
    AbstractWorkerPool: The abstract interface for all WorkerPool
        implementations.

AsyncTaskResult interface:
    This represents a Future-like object. Custom implementations of WorkerPool
    interface need to return objects with this interface in `map` method.

Exception classes:
    TaskException: raised when a worker throws during task execution.
    RekallRuntimeException: raised when there is error in the Runtime.
"""

import cloudpickle
from math import ceil
import multiprocessing as mp
import random
from tqdm import tqdm

from rekall.interval_set_3d_utils import perf_count
from rekall.domain_interval_collection import DomainIntervalCollection

class TaskException(Exception):
    """Exception to throw when a worker encounters error during task.

    Use `raise from` syntex to wrap the exception around the real error.
    For example:
        try:
            ...
        except Exception as e:
            raise TaskException() from e
    """
    def __repr__(self):
        return "TaskException from {0}".format(repr(self.__cause__))

class RekallRuntimeException(Exception):
    """Exception raised when Runtime encounters error."""
    pass

class AbstractAsyncTaskResult():
    """Definition of the AsyncTaskResult interface

    This represents a Future-like object. Custom implementations of WorkerPool
    interface need to return objects with this interface in `map` method.
    """
    def get(self):
        """Returns the value inside the AsyncTaskResult.

        This blocks until the value is ready.

        Raises:
            TaskException: if the AsyncTaskResult contains error.
        """
        raise NotImplementedError()

    def done(self):
        """Returns whether the value is ready"""
        raise NotImplementedError()

class _FutureWrapper(AbstractAsyncTaskResult):
    """Wraps a mp.pool.AsyncResult object to throw TaskException."""
    def __init__(self, future):
        self._f = future

    def get(self):
        try:
            return self._f.get()
        except Exception as e:
            raise TaskException() from e

    def done(self):
        return self._f.ready()

class AbstractWorkerPool():
    """Definition of the WorkerPool interface
    
    Notes:
        A WorkerPool instance is specialized to running one function, which is
        why the function to execute is not here in the interface but is instead
        passed to the worker pool factory function.
    """
    def map(self, tasks, done):
        """Maps the tasks over the available workers in the pool

        Args:
            tasks: A list of tasks to execute. Each task is a set of arguments
                to run the function with. 
            done: A callback function that is called when any task finishes.
                It takes the set of arguments for the finished task, and
                optionally an error that the task encountered if there is one.

        Returns:
            A list of AsyncTaskResults, one for each task in tasks.
        """
        raise NotImplementedError()

    def shut_down(self):
        """Clean up the worker pool after all tasks have finished.
        
        Implementations should release any resources used by the worker pool.
        """
        raise NotImplementedError()

class InlineSingleProcessPool(AbstractWorkerPool):
    """A single-process implmentation of WorkerPool interface."""
    class _Lazy(AbstractAsyncTaskResult):
        """A wrapper that defers the execution until result is requested"""
        def __init__(self, getter, done_cb):
            self.getter = getter
            self.done_cb = done_cb
        def get(self):
            try:
                r = self.getter()
            except Exception as e:
                self.done_cb(e)
                raise TaskException() from e
            self.done_cb()
            return r
        def done(self):
            return True

    def __init__(self, fn):
        """Initializes with the function to run."""
        self.fn = fn

    def map(self, tasks, done):
        def get_callback(vids):
            def callback(e=None):
                done(vids, e)
            return callback
        def get_getter(fn, vids):
            def getter():
                return fn(vids)
            return getter
        return [InlineSingleProcessPool._Lazy(
                    get_getter(self.fn, task),
                    get_callback(task)) for task in tasks]

    def shut_down(self):
        pass

# Helper functions for creating child processes:

# We set the function to execute as a global variable on the child process
# to avoid pickling the function.
def _child_process_init(context):
    global GLOBAL_CONTEXT
    GLOBAL_CONTEXT=context

def _apply_global_context_as_function(vids):
    global GLOBAL_CONTEXT
    fn = GLOBAL_CONTEXT
    return fn(vids)

class ForkedProcessPool(AbstractWorkerPool):
    """A WorkerPool implementation using forking.
    
    The worker processes will inherit the global context from the main process
    such as global variables. However, forking a multithreaded program safely 
    is very tricky. In particular, any global thread pool object in the parent 
    process is forked but the actual threads are not available in the forked 
    child processes.
    """
    def __init__(self, fn, num_workers):
        """Initializes the instance
        Args:
            fn: The function to run in child processes.
            num_workers: Number of child processes to create.
        """
        self._pool = mp.get_context("fork").Pool(
                processes=num_workers,
                initializer=_child_process_init,
                initargs=(fn,))

    def map(self, tasks, done_callback):
        def get_success_callback(vids):
            def success(result):
                done_callback(vids)
            return success
        def get_error_callback(vids):
            def error(err):
                done_callback(vids, err)
            return error
        return [_FutureWrapper(self._pool.apply_async(
                _apply_global_context_as_function,
                args=(task,),
                callback=get_success_callback(task),
                error_callback=get_error_callback(task))) for task in tasks]

    def shut_down(self):
        self._pool.terminate()

# When spawning, arguments to initializer are pickled.
# To allow arbitrary lambdas with closure, use cloudpickle to serialize the
# function to execute
def _apply_serialized_function(serialized_func, vids):
    fn = cloudpickle.loads(serialized_func)
    return fn(vids)

class SpawnedProcessPool(AbstractWorkerPool):
    """A WorkerPool implementation using spawning.
    
    It creates worker processes by spawning new python interpreters.
    The worker processes do not inherit any context from the main process.
    In particular, they have no access to the global variables and imported
    modules in the main process.
    """
    def __init__(self, fn, num_workers, initializer=None):
        """Initializes the instance.

        Args:
            fn: The function to run in child processes.
            num_workers: Number of child processes to create.
            initializer: A function to run in the child process after it is 
                created. It can be used to set up necessary resources in the
                worker.
        """
        self._pool = mp.get_context("spawn").Pool(
                processes=num_workers,
                initializer=initializer)
        self._pickled_fn = cloudpickle.dumps(fn)

    def map(self, tasks, done_callback):
        def get_success_callback(vids):
            def success(result):
                done_callback(vids)
            return success
        def get_error_callback(vids):
            def error(err):
                done_callback(vids, err)
            return error
        return [_FutureWrapper(self._pool.apply_async(
                _apply_serialized_function,
                args=(self._pickled_fn, task),
                callback=get_success_callback(task),
                error_callback=get_error_callback(task))) for task in tasks]

    def shut_down(self):
        self._pool.terminate()

# WorkerPool Factories
def inline_pool_factory(fn):
    """Creates a InlineSingleProcessPool."""
    return InlineSingleProcessPool(fn)

def get_forked_process_pool_factory(num_workers=mp.cpu_count()):
    """Returns a factory for ForkedProcessPool.

    Args:
        num_workers (optional): Number of child processes to fork.
            Defaults to the number of CPU cores on the machine.
    
    Returns:
        A factory for ForkedProcessPool.
    """
    def factory(fn):
        return ForkedProcessPool(fn, num_workers)
    return factory

def get_spawned_process_pool_factory(num_workers=mp.cpu_count()):
    """Returns a factory for SpawnedProcessPool.

    Args:
        num_workers (optional): Number of child processes to spawn.
            Defaults to the number of CPU cores on the machine.
    
    Returns:
        A factory for SpawnedProcessPool.
    """
    def factory(fn):
        return SpawnedProcessPool(fn, num_workers)
    return factory

class _WorkerPoolContext():
    """ Wrapper class to allow `with` syntax on WorkerPools"""
    def __init__(self, pool):
        self._pool = pool

    def __enter__(self):
        return self._pool

    def __exit__(self, *args):
        self._pool.shut_down()

def _get_callback(pbar, args_with_err, print_error=True):
    """
    Returns a callback that, when called after a task finishes, updates the
    progress bar, and if there is an error, add the task to args_with_err and
    optionally prints the error to stdout.
    """
    def callback(task_args, err=None):
        if err is None:
            if pbar is not None:
                pbar.update(len(task_args))
        else:
            if print_error:
                print("Error when processing {0}:{1}".format(
                    task_args, repr(err)))
            args_with_err.extend(task_args)
    return callback

def _create_tasks(args, chunksize):
    """Splits args into tasks of `chunksize` each."""
    total = len(args)
    num_tasks = int(ceil(total/chunksize))
    tasks = []
    for task_i in range(num_tasks):
        start = chunksize*task_i
        end = min(total,start+chunksize)
        tasks.append(args[start:end])
    return tasks

def union_combiner(result1, result2):
    """Combiner that calls union method on the result."""
    return result1.union(result2)

def disjoint_domain_combiner(result1, result2):
    """ A faster combiner than union_combiner for DomainIntervalCollection.
        
    Assumes that the results are of type DomainIntervalCollection and every 
    chunk produces its own set of domain keys that are disjoint from results of 
    other chunks.

    Args:
        result1, result2 (DomainIntervalCollection): partial results from
            some chunks of total work.
    
    Returns:
        A DomainIntervalCollection that is the union of the two.
    
    Raises:
        RekallRuntimeException: Raised if results have common domain key.
    """
    d1 = result1.get_grouped_intervals()
    d2 = result2.get_grouped_intervals()
    k1,k2 = set(d1.keys()), set(d2.keys())
    if k1.isdisjoint(k2):
        return DomainIntervalCollection({**d1, **d2})
    intersection = k1 & k2
    raise RekallRuntimeException(
        "DisjointDomainCombiner used on results"
        " with overlapping domains {0}".format(intersection))

def _pop_future_to_yield(futures):
    """Polling until a future is ready with exponential backoff"""
    MAX_WAIT = 1
    MIN_WAIT = 0.001
    FACTOR = 1.5

    def grow(wait_time):
        return min(MAX_WAIT, wait_time * FACTOR)

    from time import sleep

    wait_time = MIN_WAIT
    while True:
        for i,f in enumerate(futures):
            if f.done():
                return futures.pop(i)
        sleep(wait_time)
        wait_time = grow(wait_time)

class Runtime():
    """Manages execution of function on large number of inputs.

    Given a function that can return results for a batch of inputs, and a
    potentially long list of inputs to run the function with, Runtime helps to
    divide the inputs into small chunks, also called tasks, and dispatches
    the tasks to a pool of workers created by worker_pool_factory. It also
    gracefully handles exceptions in workers and can assemble the partial
    results.

    An example function:
        def query(video_ids):
            # Gets the intervals in the input batch of videos
            frames_with_opposing_faces = ...
            # Returns a DomainIntervalCollection with video_id as domain key.
            return frames_with_opposing_faces

        # A list of 100K video_ids
        ALL_VIDEO_IDS = ...

    In the example, query(ALL_VIDEO_IDS) is not practical to run in one go.
    To get the same results, one can use Runtime in one of two ways.

    The first way is to dispatch all tasks and wait:
        # Running the query on all videos, in chunks of 5 on 16 processes.
        rt = Runtime(get_forked_process_pool_factory(num_workers=16))
        # Will block until everything finishes
        # results is a DomainIntervalCollection with all intervals found.
        results, failed_video_ids = rt.run(
            query, ALL_VIDEO_IDS, combiner=disjoint_domain_combiner,
            chunksize=5)

    The second way is to use iterator:
        # Get an iterator that yields partial results from each chunk of 5.
        rt = Runtime(get_forked_process_pool_factory(num_workers=16))
        gen = rt.get_result_iterator(query, ALL_VIDEO_IDS, chunksize=5)
        # Blocks until the first batch is done.
        # results_from_one_batch is a DomainIntervalCollection with intervals
        # found in one task (a chunk of 5 videos).
        results_from_one_batch = next(gen)
    """
    def __init__(self, worker_pool_factory):
        """Initialized with a WorkerPool Factory
        
        Args:
            worker_pool_factory: A function that takes the query to execute,
                and returns a worker pool to execute the query.
        """
        self._get_worker_pool = worker_pool_factory

    @classmethod
    def inline(cls):
        """Inline Runtime executes each chunk in sequence in one process."""
        return cls(inline_pool_factory)

    def run(self, query, args, combiner=union_combiner,
            randomize=True, chunksize=1,
            progress=False, profile=False,
            print_error=True):
        """Dispatches all tasks to workers and waits until everything finishes.

        See class documentation for an example of how to use run().
        Exception raised in `query` are suppressed and the unsuccessful subset
        of `args` is returned at the end. However, such errors can be printed
        as soon as they occur.

        Args:
            query: A function that can return partial results for any batch of
                input arguments.
            args: A potentially long list of input arguments to execute the
                query with.
            combiner (optional): A function that takes two partial results and
                returns the combination of the two.
                Defaults to union_combiner which assumes the partial results
                have a `union` method.
            randomize (optional): Whether to create and dispatch tasks in
                random order.
                Defaults to True.
            chunksize (optional): The size of the input batch for each task.
                Defaults to 1.
            progress (optional): Whether to display a progress bar.
                Defaults to False.
            profile (optional): Whether to output wall time of various internal
                stages to stdout.
                Defaults to False.
            print_error (optional): Whether to output task errors to stdout.
                Defaults to True.

        Returns:
            A pair (query_output, args_with_err) where
            query_output: The combined results from successful tasks.
            args_with_err: A list that is a subset of args that failed to
                execute.
        """
        with perf_count("Executing query in Runtime", enable=profile):
            with _WorkerPoolContext(self._get_worker_pool(query)) as pool:
                total_work = len(args)
                with tqdm(total=total_work, disable=not progress) as pbar:
                    with perf_count("Executing in workers", enable=profile):
                        args_with_err = []
                        with perf_count("Dispatching tasks", enable=profile):
                            if randomize:
                                random.shuffle(args)
                            async_results = pool.map(
                                    _create_tasks(args, chunksize),
                                    _get_callback(pbar, args_with_err,
                                        print_error))
                        combined_result = None
                        for future in async_results:
                            try:
                                r = future.get()
                            except TaskException:
                                continue
                            if combined_result is None:
                                combined_result = r
                            else:
                                combined_result = combiner(combined_result, r)
                        if combined_result is None and total_work>0:
                            raise RekallRuntimeException("All tasks failed!")
                        return (combined_result, args_with_err)

    def get_result_iterator(self, query, args, randomize=True, chunksize=1, 
            print_error=True, dispatch_size=mp.cpu_count()):
        """Incrementally dispatches tasks as partial results are consumed.

        See class documentation for an example of how to use 
        get_result_iterator().
        Exception raised in `query` are suppressed and if any tasks failed,
        will raise a RekallRuntimeException after all successful tasks' results
        have been yielded. However, such errors can be printed as soon as they
        occur.

        Args:
            query, args, randomize, chunksize, print_error: Same as in run().
            dispatch_size (int, optional): Number of tasks to dispatch at a
                time. In this mode, tasks are incrementally dispatched
                as partial results from preivous tasks are yielded.
                If not positive, will dispatch all tasks at once.
                Defaults to the number of CPU cores.

        Yields:
            Partial results from each task.

        Raises:
            RekallRuntimeException: Raised after all successful task results
                have been yielded if there have been failed tasks.
        """
        with _WorkerPoolContext(self._get_worker_pool(query)) as pool:
            args_with_err = []
            if randomize:
                random.shuffle(args)
            tasks = _create_tasks(args, chunksize)
            if dispatch_size is None or dispatch_size<=0:
                dispatch_size = len(tasks)
            outstanding_tasks = tasks
            async_results = []
            num_finished_tasks = 0
            while num_finished_tasks < len(tasks):
                # Maybe make a dispatch
                num_to_yield = len(async_results)
                if (num_to_yield <= dispatch_size/2 and 
                    len(outstanding_tasks) > 0):
                    task_batch = outstanding_tasks[:dispatch_size]
                    outstanding_tasks = outstanding_tasks[dispatch_size:]
                    async_results.extend(pool.map(
                        task_batch,
                        _get_callback(None, args_with_err, print_error)))
                if randomize:
                    future_to_yield = _pop_future_to_yield(async_results)
                else:
                    future_to_yield = async_results.pop(0)
                num_finished_tasks += 1
                try:
                    r = future_to_yield.get()
                except TaskException:
                    continue
                yield r
        if len(args_with_err) > 0:
            raise RekallRuntimeException(
                    "The following tasks failed: {0}".format(
                args_with_err))




from rekall.logical_predicates import *
from rekall.temporal_predicates import *
from rekall.interval_list import Interval
import unittest

class TestLogicalPredicates(unittest.TestCase):
    def test_true_pred(self):
        intrvl1 = Interval(1, 2, 1)
        intrvl2 = Interval(5, 6, 1)
        self.assertTrue(true_pred(arity=2)(intrvl1, intrvl2))

    def test_false_pred(self):
        intrvl1 = Interval(1, 2, 1)
        intrvl2 = Interval(5, 6, 1)
        self.assertFalse(false_pred(arity=2)(intrvl1, intrvl2))

    def test_and(self):
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(1., 2., 3)
        self.assertTrue(and_pred(equal(), overlaps(), arity=2)(intrvl1, intrvl2))
        self.assertFalse(and_pred(equal(), overlaps_before(), arity=2)(intrvl1, intrvl2))
        self.assertFalse(and_pred(overlaps_before(), equal(), arity=2)(intrvl1, intrvl2))
        self.assertFalse(and_pred(overlaps_before(), overlaps_after(), arity=2)(intrvl1, intrvl2))

    def test_or(self):
        intrvl1 = Interval(1., 3., 1)
        intrvl2 = Interval(2., 4., 1)

        self.assertTrue(or_pred(before(), overlaps_before(), arity=2)(intrvl1, intrvl2))
        self.assertFalse(or_pred(before(), false_pred(arity=2), arity=2)(intrvl1, intrvl2))
        self.assertTrue(or_pred(overlaps_before(), overlaps(), arity=2)(intrvl1, intrvl2))
        self.assertTrue(or_pred(overlaps_before(), before(), arity=2)(intrvl1, intrvl2))

    def test_not(self):
        intrvl1 = Interval(1., 3., 1)
        intrvl2 = Interval(2., 4., 1)

        self.assertTrue(overlaps()(intrvl1, intrvl2))
        self.assertFalse(not_pred(overlaps(), arity=2)(intrvl1, intrvl2))




from setuptools import setup

if __name__ == "__main__":
    setup(
        name='rekallpy',
        version='0.1.0',
        description='Spatiotemporal query language',
        url='http://github.com/scanner-research/rekall',
        author='Dan Fu',
        author_email='danfu@stanford.edu',
        license='Apache 2.0',
        packages=['rekall'],
        install_requires=['python-constraint', 'tqdm', 'cloudpickle'],
        setup_requires=['pytest-runner'],
        tests_require=['pytest'],
        zip_safe=False)



"""Binary predicates on Temporal Ranges.

Before and After:
    These s optionally take a min_dist and max_dist. They check if
    the distance between intrvl1 and intrvl2 is in the range
    [min_dist, max_dist] (in the right direction). Note that by default, this
    includes intervals that meet each other.

OverlapsBefore and OverlapsAfter:
    The strict Allen interval definition of overlapping in either direction.
    Returns true if intrvl1 and intrvl2 have the following relationship:
      
      |-----|
         |-----|
    
    OverlapsAfter requires that intrvl1 start after intrvl2 (and end after
    intrvl2).

Starts and StartsInv:
    True if intrvl1 has same start time as intrvl2 and ends before intrvl2 
    (flip intrvl1 and intrvl2 for the inverse).

Finishes and FinishesInv:
    True if intrvl1 has the same finish time as intrvl2 and starts before 
    intrvl2 (flip for inverse).

During and DuringInv:
    True if intrvl1 starts sintrvlictly after intrvl2 and ends strictly before
    intrvl2 (flip for inverse).

MeetsBefore and MeetsAfter:
    True if intrvl1 starts when intrvl2 ends (flip for inverse).

Equal:
    True if intrvl1 and intrvl2 start and end at the same time.

Overlaps:
    Sugar for a more colloquial version of overlapping. Includes Starts/Inv,
    Finishes/Inv, During/Inv, Equal, and OverlapsBefore/After.
"""

from rekall.common import *

def before(min_dist=0, max_dist=INFTY):
    def fn(intrvl1, intrvl2):
        time_diff = intrvl2.start - intrvl1.end
        return (time_diff >= min_dist and
            (max_dist == INFTY or time_diff <= max_dist))

    return fn

def after(min_dist=0, max_dist=INFTY):
    def fn(intrvl1, intrvl2):
        time_diff = intrvl1.start - intrvl2.end
        return (time_diff >= min_dist and
            (max_dist == INFTY or time_diff <= max_dist))

    return fn

def overlaps():
    return lambda intrvl1, intrvl2: ((intrvl1.start < intrvl2.start and intrvl1.end > intrvl2.start) or
            (intrvl1.start < intrvl2.end and intrvl1.end > intrvl2.end) or
            (intrvl1.start <= intrvl2.start and intrvl1.end >= intrvl2.end) or
            (intrvl1.start >= intrvl2.start and intrvl1.end <= intrvl2.end))

def overlaps_before():
    return lambda intrvl1, intrvl2: (intrvl1.end > intrvl2.start and intrvl1.end < intrvl2.end and
            intrvl1.start < intrvl2.start)

def overlaps_after():
    return lambda intrvl1, intrvl2: (intrvl1.start > intrvl2.start and intrvl1.start < intrvl2.end and
            intrvl1.end > intrvl2.end)

def starts(epsilon=0):
    return lambda intrvl1, intrvl2: (abs(intrvl1.start - intrvl2.start) <= epsilon
            and intrvl1.end < intrvl2.end)

def starts_inv(epsilon=0):
    return lambda intrvl1, intrvl2: (abs(intrvl1.start - intrvl2.start) <= epsilon
            and intrvl2.end < intrvl1.end)

def finishes(epsilon=0):
    return lambda intrvl1, intrvl2: (abs(intrvl1.end - intrvl2.end) <= epsilon
            and intrvl1.start > intrvl2.start)

def finishes_inv(epsilon=0):
    return lambda intrvl1, intrvl2: (abs(intrvl1.end - intrvl2.end) <= epsilon
            and intrvl2.start > intrvl1.start)

def during():
    return lambda intrvl1, intrvl2: intrvl1.start > intrvl2.start and intrvl1.end < intrvl2.end

def during_inv():
    return lambda intrvl1, intrvl2: intrvl2.start > intrvl1.start and intrvl2.end < intrvl1.end

def meets_before(epsilon=0):
    return lambda intrvl1, intrvl2: abs(intrvl1.end-intrvl2.start) <= epsilon

def meets_after(epsilon=0):
    return lambda intrvl1, intrvl2: abs(intrvl2.end-intrvl1.start) <= epsilon

def equal():
    return lambda intrvl1, intrvl2: intrvl1.start == intrvl2.start and intrvl1.end == intrvl2.end




"""Logical predicates

They take in an optional arity parameter.
"""

from rekall.helpers import *

def and_pred(pred1, pred2, arity=1):
    if arity == 1:
        return lambda intrvl: pred1(intrvl) and pred2(intrvl)
    elif arity == 2:
        return lambda intrvl1, intrvl2: (pred1(intrvl1, intrvl2) and
                pred2(intrvl1, intrvl2))
    else:
        panic('arity {} not supported.'.format(arity))

def or_pred(pred1, pred2, arity=1):
    if arity == 1:
        return lambda intrvl: pred1(intrvl) or pred2(intrvl)
    elif arity == 2:
        return lambda intrvl1, intrvl2: (pred1(intrvl1, intrvl2) or
                pred2(intrvl1, intrvl2))
    else:
        panic('arity {} not supported.'.format(arity))

def true_pred(arity=1):
    if arity == 1:
        return lambda intrvl: True
    elif arity == 2:
        return lambda intrvl1, intrvl2: True
    else:
        panic('arity {} not supported.'.format(arity))

def false_pred(arity=1):
    if arity == 1:
        return lambda intrvl: False
    elif arity == 2:
        return lambda intrvl1, intrvl2: False
    else:
        panic('arity {} not supported.'.format(arity))

def not_pred(pred, arity=1):
    if arity == 1:
        return lambda intrvl: not pred(intrvl)
    elif arity == 2:
        return lambda intrvl1, intrvl2: not pred(intrvl1, intrvl2)
    else:
        panic('arity {} not supported.'.format(arity))



"""Predicates on payloads.

Right now just two predicates that takes in an interval
and a predicate and applies the payload to the predicate (for both binary and
unary predicates):
    payload_satisfies(pred, 1) - True if the interval's payload satisfies
        pred
    payload_staisfies(pred, 2) - True if the two interval's payloads
        satisfy pred
To apply pred on a named payload, use payload_satisfies(on_name(name, pred))
"""
from rekall.helpers import *

def payload_satisfies(pred, arity=1):
    if arity == 1:
        return lambda intrvl: pred(intrvl.payload)
    elif arity == 2:
        return lambda intrvl1, intrvl2: pred(intrvl1.payload, intrvl2.payload)
    else:
        panic('arity {} not supported.'.format(arity))

def on_name(name, pred):
    """
    Wraps a predicate so that it can be applied to one value in a dict payload
    """
    def apply(*args):
        return pred(*[arg[name] for arg in args])
    return apply



from rekall.interval_set_3d import Interval3D, IntervalSet3D
import rekall.interval_set_3d_utils as utils
from rekall.merge_ops import payload_plus, payload_second
from rekall.temporal_predicates import (overlaps, before, meets_before, equal,
    during_inv)
    
from rekall.bbox_predicates import left_of, above, height_at_least
import unittest
from operator import eq
import json

class TestInterval3D(unittest.TestCase):
    @staticmethod
    def intervalsEq(i1, i2, payload_cmp=None):
        def float_equal(f1, f2):
            return abs(f1-f2)<=1e-7
        def bound_almost_eq(b1,b2):
            return (float_equal(b1[0], b2[0]) and
                   float_equal(b1[1], b2[1]))
            
        return (i1.t==i2.t and bound_almost_eq(i1.x, i2.x) and 
                bound_almost_eq(i1.y,i2.y) and (
                payload_cmp is None or payload_cmp(i1.payload,i2.payload)))

    def assertIntervalEq(self, i1, i2, payload_cmp=None):
        self.assertTrue(TestInterval3D.intervalsEq(i1,i2,payload_cmp),
                msg = "Interval {0} different from Interval {1}".format(
                    repr(i1), repr(i2)))

    def test_merge(self):
        i1 = Interval3D((0,1),(0,0.5),(0,0.5),1)
        i2 = Interval3D((1,2),(0.4,1),(0.2,0.3),2)
        i3 = i1.merge(i2, payload_plus)
        target = Interval3D((0,2),(0,1), (0,0.5), 3)
        self.assertIntervalEq(i3, target, eq)

    def test_combine(self):
        def combiner1(b1, b2):
            return b1
        def combiner2(b1, b2):
            return b2
        i1 = Interval3D((0,1),(0,0.5),(0,0.1))
        i2 = Interval3D((2,3),(2,2.5),(1,1.1))
        i3 = i1.combine(i2, combiner2, combiner1, combiner2)
        target = Interval3D((2,3),(0,0.5),(1,1.1))
        self.assertIntervalEq(i3, target)

    def test_overlap_time_merge_space(self):
        i1 = Interval3D((0,2.5),(0,0.5),(0,0.1))
        i2 = Interval3D((2,3),(2,2.5),(1,1.1))
        i3 = i1.overlap_time_merge_space(i2)
        target = Interval3D((2,2.5),(0,2.5), (0,1.1))
        self.assertIntervalEq(i3, target)

    def test_overlap_time_merge_space_throws(self):
        i1 = Interval3D((0,1),(0,0.5),(0,0.1))
        i2 = Interval3D((2,3),(2,2.5),(1,1.1))
        with self.assertRaises(ValueError):
            i3 = i1.overlap_time_merge_space(i2)

    def test_expand_to_frame(self):
        i1 = Interval3D((0,1),(0.5,0.6),(0.7,0.8)).expand_to_frame()
        self.assertIntervalEq(i1,
                Interval3D((0,1),(0,1),(0,1)))

    def test_dimension_sizes(self):
        i = Interval3D((0,2),(0.5,0.8),(0.9,1.0))
        self.assertAlmostEqual(i.length(), 2)
        self.assertAlmostEqual(i.width(), 0.3)
        self.assertAlmostEqual(i.height(), 0.1)
        
class TestIntervalSet3D(unittest.TestCase):
    def assertIntervalSetEq(self, is1, is2, payload_cmp=None):
        list1 = is1.get_intervals()
        list2 = is2.get_intervals()
        for i1, i2 in zip(list1, list2):
            self.assertTrue(TestInterval3D.intervalsEq(i1,i2,payload_cmp),
                    "Interval {0} different from Interval {1}".format(
                        repr(i1), repr(i2)))
        self.assertFalse(len(list1)<len(list2),
                "2nd IntervalSet has more intervals {0}".format(
                    list2[len(list1):]))
        self.assertFalse(len(list1)>len(list2),
                "1st IntervalSet has more intervals {0}".format(
                    list1[len(list2):]))

    def compare_interval_sets_in_payload(self):
        def payload_cmp(set1, set2):
            # Will throw if not equal
            self.assertIntervalSetEq(set1, set2)
            return True
        return payload_cmp

    def test_map(self):
        is1 = IntervalSet3D([
            Interval3D((0,1),(0.3,0.4),(0.5,0.6)),
            Interval3D((0,0.5),(0.2,0.3),(0.5,0.6)),
            ])
        is1 = is1.map(Interval3D.expand_to_frame)
        target = IntervalSet3D([
            Interval3D((0,0.5),(0,1),(0,1)),
            Interval3D((0,1),(0,1),(0,1)),
            ])
        self.assertIntervalSetEq(is1, target)

    def test_union(self):
        is1 = IntervalSet3D([
            Interval3D((0,1),(0,1),(0,1),1),
            Interval3D((0,0.5),(0.5,1),(0,0.5),1),
            ])
        is2 = IntervalSet3D([
            Interval3D((0.5,1),(0,1),(0,1),2),
            Interval3D((0,1),(0,1),(0,1),2),
            ])
        is3 = is1.union(is2)
        target = IntervalSet3D([
            Interval3D((0,1),(0,1),(0,1),1),
            Interval3D((0,1),(0,1),(0,1),2),
            Interval3D((0,0.5),(0.5,1),(0,0.5),1),
            Interval3D((0.5,1),(0,1),(0,1),2),
            ])
        self.assertIntervalSetEq(is3, target, eq)

    def test_join(self):
        is1 = IntervalSet3D([
            Interval3D((0,1),(0,1),(0,1),1),
            Interval3D((0,0.5),(0.5,1),(0,0.5),2),
            ])
        is2 = IntervalSet3D([
            Interval3D((0.5,1),(0,1),(0,1),4),
            Interval3D((0,1),(0,1),(0,1),8),
            ])
        is3 = is1.join(is2,
                utils.T(overlaps()),
                lambda i1,i2: [i1.overlap_time_merge_space(i2,
                    payload_plus)])
        target = IntervalSet3D([
            Interval3D((0.5,1),(0,1),(0,1),5),
            Interval3D((0,1),(0,1),(0,1),9),
            Interval3D((0,0.5),(0,1),(0,1),10),
            ])
        self.assertIntervalSetEq(is3, target, eq)

    def test_join_with_time_window(self):
        is1 = IntervalSet3D([
            Interval3D((t,t+1), payload=t) for t in range(100)
            ])
        is2 = IntervalSet3D([
            Interval3D((t,t), payload=t) for t in range(100)
            ])
        is3 = is1.join(is2,
                utils.T(before(max_dist=1)),
                lambda i1, i2: [i1.merge(i2, payload_second)],
                time_window=1)
        target = IntervalSet3D([
            Interval3D((t, t+2), payload=t+2) for t in range(98)]+[
            Interval3D((t,t+1), payload=t+1) for t in range(99)])
        self.assertIntervalSetEq(is3, target, eq)

    def test_filter(self):
        is1 = IntervalSet3D([
            Interval3D((0,1),(0,1),(0,1),1),
            Interval3D((0,0.5),(0.5,1),(0,0.5),2),
            Interval3D((0,0.5),(0,1),(0,0.5),3),
            ])
        is1 = is1.filter(lambda i: i.payload > 2)
        target = IntervalSet3D([
            Interval3D((0,0.5),(0,1),(0,0.5),3),
            ])
        self.assertIntervalSetEq(is1, target, eq)

    def test_fold(self):
        def fold_fn(acc, i):
            return acc + i.length()*i.width()*i.height()

        is1 = IntervalSet3D([
            Interval3D((0,1),(0,1),(0,1),1),
            Interval3D((0,0.5),(0.5,1),(0,0.5),2),
            Interval3D((0,0.5),(0,1),(0,0.5),3),
            ])
        self.assertAlmostEqual(is1.fold(fold_fn, 0),
                1.375)

    def test_fold_custom_sortkey(self):
        def sortkey(i):
            return i.x
        def fold_fn(acc, i):
            acc.append(i.payload)
            return acc
        is1 = IntervalSet3D([
            Interval3D((0,1),(0,0.1),(0,1),1),
            Interval3D((0,0.5),(0.5,1),(0,0.5),2),
            Interval3D((0.1,0.4),(0,1),(0,0.5),3),
            ])
        self.assertListEqual(is1.fold(fold_fn, [], sortkey),
                [1,3,2])

    def test_group_by(self):
        is1 = IntervalSet3D([
            Interval3D((0,1)),
            Interval3D((1,2)),
            Interval3D((0,2)),
            Interval3D((1,3)),
            ])

        def merge_intervals(k, intervals):
            merged = intervals.fold(Interval3D.merge)
            merged.payload = intervals
            return merged

        is2 = is1.group_by(lambda i: int(i.t[1]) % 2,
                merge_intervals)
        target = IntervalSet3D([
            Interval3D((0,3), payload = IntervalSet3D([
                Interval3D((0,1)),
                Interval3D((1,3)),])),
            Interval3D((0,2), payload = IntervalSet3D([
                Interval3D((0,2)),
                Interval3D((1,2)),])),
            ])
        self.assertIntervalSetEq(is2, target,
                self.compare_interval_sets_in_payload())

    def test_minus(self):
        is1 = IntervalSet3D([
            Interval3D((1, 10),(0,0.5),(0.2,0.8),1),
            Interval3D((3, 15),(0,1),(0,1),2)])
        is2 = IntervalSet3D([
            Interval3D((2,2.5)),
            Interval3D((2,2.7)),
            Interval3D((2.9,3.5)),
            Interval3D((3.5,3.6)),
            Interval3D((5,7)),
            Interval3D((9,12)),
            ])
        is3 = is1.minus(is2)
        target = IntervalSet3D([
            Interval3D((1,2),(0,0.5),(0.2,0.8),1),
            Interval3D((2.7, 2.9), (0,0.5),(0.2,0.8),1),
            Interval3D((3.6,5),(0,0.5),(0.2,0.8),1),
            Interval3D((7,9),(0,0.5),(0.2,0.8),1),
            Interval3D((3.6,5), payload=2),
            Interval3D((7,9), payload=2),
            Interval3D((12,15), payload=2),
            ])
        self.assertIntervalSetEq(is3, target, eq)

    def test_minus_everything(self):
        is1 = IntervalSet3D([
            Interval3D((1, 10)),
            Interval3D((3, 15))])
        is2 = IntervalSet3D([
            Interval3D((2,2.5)),
            Interval3D((2,2.7)),
            Interval3D((2.9,3.5)),
            Interval3D((3.5,3.6)),
            Interval3D((5,7)),
            Interval3D((9,12)),
            ])
        is3 = is2.minus(is1)
        self.assertIntervalSetEq(is3, IntervalSet3D([]), eq)

    def test_minus_self(self):
        is1 = IntervalSet3D([
            Interval3D((2,2.5)),
            Interval3D((2,2.7)),
            Interval3D((2.9,3.5)),
            Interval3D((3.5,3.6)),
            Interval3D((5,7)),
            Interval3D((9,12)),
            ])
        is1 = is1.minus(is1)
        self.assertIntervalSetEq(is1, IntervalSet3D([]), eq)

    def test_minus_against_nothing(self):
        is1 = IntervalSet3D([
            Interval3D((1, 10),(0,0.5),(0.2,0.8),1),
            Interval3D((3, 15),(0,1),(0,1),2)])
        is2 = IntervalSet3D([
            Interval3D((20,20.5)),
            Interval3D((20,20.7)),
            Interval3D((20.9,23.5)),
            Interval3D((23.5,23.6)),
            Interval3D((25,27)),
            Interval3D((29,32)),
            ])
        is3 = is1.minus(is2)
        self.assertIntervalSetEq(is3, is1, eq)

    def test_minus_with_single_frame(self):
        is1 = IntervalSet3D([
            Interval3D((1,1)),
            Interval3D((3,3)),
            Interval3D((4,4)),
            Interval3D((7,7)),
            Interval3D((10,10)),
            ])
        is2 = IntervalSet3D([
            Interval3D((1,3)),
            Interval3D((5,8)),
            Interval3D((9,9)),
            ])
        is3 = is1.minus(is2)
        target = IntervalSet3D([
            Interval3D((4,4)),
            Interval3D((10,10)),
            ])
        self.assertIntervalSetEq(is3, target)

        is4 = is2.minus(is1)
        self.assertIntervalSetEq(is4, is2)

    def test_match(self):
        left_box_frame_1 = Interval3D((1,1),(0.1,0.4),(0.4,0.8))
        right_box_frame_1 = Interval3D((1,1),(0.6,0.9),(0.3,0.7))
        bottom_left_box_frame_2 = Interval3D((2,2),(0.1,0.3),(0.8,0.9))
        top_right_box_frame_2 = Interval3D((2,2),(0.5,0.7),(0.2,0.7))
        is1 = IntervalSet3D([
            left_box_frame_1,
            right_box_frame_1,
            bottom_left_box_frame_2,
            top_right_box_frame_2])

        pattern = [
                (["left","right"],[
                    # Two boxes on left and right on same frame
                    utils.T(equal()),
                    utils.XY(left_of()),
                ]),
                (["top", "bottom"], [
                    # Two boxes on top and bottom on overlapping frame
                    utils.T(equal()),
                    utils.XY(above()),
                ]),
                (["left", "top"],[
                    # Left-Right pattern comes before Top-Bottom
                    utils.T(meets_before(epsilon=1))
                ])
        ]
        results = is1.match(pattern, exact=True)
        self.assertEqual(len(results),1)
        result = results[0]
        self.assertTrue(TestInterval3D.intervalsEq(
            left_box_frame_1, result['left']))
        self.assertTrue(TestInterval3D.intervalsEq(
            right_box_frame_1, result['right']))
        self.assertTrue(TestInterval3D.intervalsEq(
            top_right_box_frame_2, result['top']))
        self.assertTrue(TestInterval3D.intervalsEq(
            bottom_left_box_frame_2, result['bottom']))

    def test_match_multiple_solutions(self):
        left_box_frame_1 = Interval3D((1,1),(0.1,0.4),(0.4,0.8))
        right_box_frame_1 = Interval3D((1,1),(0.6,0.9),(0.3,0.7))
        bottom_left_box_frame_2 = Interval3D((2,2),(0.1,0.3),(0.8,0.9))
        top_right_box_frame_2 = Interval3D((2,2),(0.5,0.7),(0.2,0.7))
        is1 = IntervalSet3D([
            left_box_frame_1,
            right_box_frame_1,
            bottom_left_box_frame_2,
            top_right_box_frame_2,
            Interval3D((3,3)),
            ])

        pattern = [
                (["left","right"],[
                    # Two boxes on left and right on same frame
                    utils.T(equal()),
                    utils.XY(left_of()),
                ]),
        ]
        results = is1.match(pattern, exact=False)
        self.assertEqual(len(results),2)

        # Add single interval constraints.
        pattern = pattern + [
                (["left"], [utils.XY(height_at_least(0.3))]),
                (["right"], [utils.XY(height_at_least(0.3))]),
                ]
        results = is1.match(pattern, exact=False)
        self.assertEqual(len(results),1)
        result = results[0]
        self.assertTrue(TestInterval3D.intervalsEq(
            left_box_frame_1, result['left']))
        self.assertTrue(TestInterval3D.intervalsEq(
            right_box_frame_1, result['right']))

    def test_filter_against(self):
        is1 = IntervalSet3D([
            Interval3D((0,1)),
            Interval3D((2,5)),
            Interval3D((6,7)),
            ])
        is2 = IntervalSet3D([
            Interval3D((1,1)),
            Interval3D((7,7)),
            ])
        # Take only intervals in is1 that overlaps with some interval in is2
        is3 = is1.filter_against(is2, utils.T(overlaps()),
                time_window=0)
        self.assertIntervalSetEq(is3,
                IntervalSet3D([Interval3D((0,1)),Interval3D((6,7))]))

    def test_map_payload(self):
        is1 = IntervalSet3D([
            Interval3D((0,1),payload=1),
            Interval3D((2,3),payload=2),
            Interval3D((4,5),payload=3),
            ])
        is2 = is1.map_payload(lambda p: p+10)
        target = IntervalSet3D([
            Interval3D((0,1),payload=11),
            Interval3D((2,3),payload=12),
            Interval3D((4,5),payload=13),
            ])
        self.assertIntervalSetEq(is2, target)

    def test_dilate(self):
        is1 = IntervalSet3D([
            Interval3D((2,5)),
            Interval3D((6,7)),
        ])
        is2 = is1.dilate(1)
        target = IntervalSet3D([
            Interval3D((1,6)),
            Interval3D((5,8)),
        ])
        self.assertIntervalSetEq(is2, target)
        self.assertIntervalSetEq(is2.dilate(-1), is1)
        is3 = is1.dilate(-0.1, axis='X')
        target = IntervalSet3D([
            Interval3D((2,5),(0.1, 0.9),(0,1)),
            Interval3D((6,7),(0.1, 0.9),(0,1)),
            ])
        self.assertIntervalSetEq(is3, target)

    def test_filter_size(self):
        is1 = IntervalSet3D([
            Interval3D((1,2),(0.5,0.9),(0.1,0.2)),
            Interval3D((20,30),(0.4,1.0), (0,1)),
            Interval3D((50,55),(0.2,0.3), (0.5,0.9)),
            ])
        is2 = is1.filter_size(min_size=10, axis='T')
        self.assertIntervalSetEq(is2, IntervalSet3D([
            Interval3D((20,30),(0.4,1.0),(0,1))]))

        is3 = is1.filter_size(max_size=0.1, axis='Y')
        self.assertIntervalSetEq(is3, IntervalSet3D([
            Interval3D((1,2),(0.5,0.9),(0.1,0.2))]))

        is4 = is1.filter_size(min_size=0.3, max_size=0.5, axis='X')
        self.assertIntervalSetEq(is3, IntervalSet3D([
            Interval3D((1,2),(0.5,0.9),(0.1,0.2))]))

    def test_merge(self):
        is1 = IntervalSet3D([
            Interval3D((1,10),(0.8,0.9),(0,0.1)),
            Interval3D((20,30),(0.1,0.2),(0.9,1)),
            ])
        is2 = IntervalSet3D([
            Interval3D((11,15),(0.4,0.5),(0.8,0.9)),
            Interval3D((35,36),(0,1),(0,1)),
            ])
        is3 = is1.merge(is2, utils.T(meets_before(epsilon=1)))
        target = IntervalSet3D([
            Interval3D((1,15),(0.4,0.9),(0,0.9)),
            ])
        self.assertIntervalSetEq(is3, target)

    def test_group_by_time(self):
        intervals_1 = [
            Interval3D((1,1), (0.4,0.5),(0.6,0.8), 1),
            Interval3D((1,1), (0.1,0.2),(0.2,0.3), 2),
            Interval3D((1,1), (0.3,0.5),(0.1,0.5), 3),
        ]
        intervals_2 = [
            Interval3D((2,2), (0.3,0.5),(0.6,0.8), 11),
            Interval3D((2,2), (0.2,0.3),(0.2,0.9), 12),
            Interval3D((2,2), (0.3,0.7),(0,0.5), 13),
        ]
        is1 = IntervalSet3D(intervals_1+intervals_2)
        target = IntervalSet3D([
            Interval3D((1,1), payload=IntervalSet3D(intervals_1)),
            Interval3D((2,2), payload=IntervalSet3D(intervals_2)),
            ])
        self.assertIntervalSetEq(is1.group_by_time(), target,
                self.compare_interval_sets_in_payload())

    def test_collect_by_interval(self):
        is1 = IntervalSet3D([
            Interval3D((1,5)),
            Interval3D((10,50)),
            Interval3D((51,52)),
            ])
        intervals = [
            Interval3D((2,2), (0.4,0.6),(0.1,0.2), 1),
            Interval3D((3,3), (0.1,0.6),(0.8,0.9), 2),
            Interval3D((11,23), (0,1),(0.8,0.9), 3),
            ]
        is2 = IntervalSet3D(intervals)

        is3 = is1.collect_by_interval(is2, utils.T(during_inv()),
                filter_empty=True, time_window=None).map_payload(lambda p:p[1])
        target = IntervalSet3D([
            Interval3D((1,5), payload=IntervalSet3D(intervals[:2])),
            Interval3D((10,50), payload=IntervalSet3D(intervals[2:3])),
            ])
        self.assertIntervalSetEq(is3, target,
                self.compare_interval_sets_in_payload())

        is4 = is1.collect_by_interval(is2, utils.T(overlaps()),
                filter_empty=False, time_window=0).map_payload(lambda p:p[1])
        target = IntervalSet3D([
            Interval3D((1,5), payload=IntervalSet3D(intervals[:2])),
            Interval3D((10,50), payload=IntervalSet3D(intervals[2:3])),
            Interval3D((51,52), payload=IntervalSet3D([])),
            ])
        self.assertIntervalSetEq(is4, target,
                self.compare_interval_sets_in_payload())

    def test_temporal_coalesce(self):
        is1 = IntervalSet3D([
            Interval3D((1,10),(0.3,0.4),(0.5,0.6),1),
            Interval3D((2,5),(0.2,0.8),(0.2,0.3),1),
            Interval3D((10,11),(0.2,0.7),(0.3,0.5),1),
            Interval3D((13,15),(0.5,1),(0,0.5),1),
            Interval3D((15,19),(0.5,1),(0,0.5),1),
            Interval3D((20,20),payload=1),
            Interval3D((22,22),payload=1),
            Interval3D((22,23),payload=1),
            ])
        target = IntervalSet3D([
            Interval3D((1,11),(0.2,0.8),(0.2,0.6),3),
            Interval3D((13,19),(0.5,1),(0,0.5),2),
            Interval3D((20,20),payload=1),
            Interval3D((22,23),payload=2),
            ])
        self.assertIntervalSetEq(is1.temporal_coalesce(payload_plus), target)
        self.assertIntervalSetEq(
                is1.temporal_coalesce(payload_plus, epsilon=2),
                IntervalSet3D([Interval3D((1,23),payload=8)]))

    def test_split(self):
        is1 = IntervalSet3D([
            Interval3D((1,5)),
            Interval3D((10,22)),
            ])
        def split_fn(i):
            output = []
            t = i.copy()
            while t.length() > 5:
                output.append(Interval3D((t.t[0],t.t[0]+5)))
                t.t = (t.t[0]+5, t.t[1])
            if t.length() > 0:
                output.append(t)
            return IntervalSet3D(output)
        target = IntervalSet3D([
            Interval3D((1,5)),
            Interval3D((10,15)),
            Interval3D((15,20)),
            Interval3D((20,22)),
            ])
        self.assertIntervalSetEq(target, is1.split(split_fn))

    def test_len(self):
        is1 = IntervalSet3D([
            Interval3D((1,5)),
            Interval3D((10,22)),
            ])
        self.assertEqual(len(is1), 2)
        self.assertEqual(is1.size(), 2)

    def test_json_conversion(self):
        is1 = IntervalSet3D([
            Interval3D((1,5), payload=5),
            Interval3D((10,50),(0,1),(0.5,1),
                payload=IntervalSet3D([
                    Interval3D((20,25))
                    ]))
            ])
        self.assertEqual(
                is1.to_json(),
                [{
                    "t": (1,5),
                    "x": (0,1),
                    "y": (0,1),
                    "payload": 5
                },{
                    "t": (10,50),
                    "x": (0,1),
                    "y": (0.5,1),
                    "payload": [
                        {"t": (20,25), "x":(0,1), "y":(0,1), "payload":None}
                    ]
                }])
